# Copyright 2019-2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Tuple

from aioredlock import Aioredlock, Lock, LockAcquiringError, LockError
from structlog import get_logger

logger = get_logger(__name__)


class RedisDistLockManager:
    """Create Distributed Locks in Redis.

    https://redis.io/topics/distlock
    """

    namespace = "orchestrator:distlock"

    def __init__(self, redis_address: Tuple[str, int]):
        self.redis_distlock: Optional[Aioredlock] = None
        self.redis_address = redis_address

    async def connect_redis(self) -> None:
        self.redis_distlock = Aioredlock([self.redis_address])

    async def disconnect_redis(self) -> None:
        if self.redis_distlock:
            await self.redis_distlock.destroy()

    async def get_lock(self, resource: str, expiration_seconds: int) -> Optional[Lock]:
        if not self.redis_distlock:
            return None

        key = f"{self.namespace}:{resource}"
        try:
            return await self.redis_distlock.lock(key, expiration_seconds)
        except LockAcquiringError:
            # normal behavior (lock acquired by something else)
            return None
        except LockError:
            # Unexpected behavior, possibly a problem with Redis
            logger.Exception("Could not acquire lock for resource", resource=key)
            return None

    async def release_lock(self, lock: Lock) -> None:
        if not self.redis_distlock:
            return None

        try:
            await self.redis_distlock.unlock(lock)
        except LockError:
            logger.Exception("Could not release lock for resource", resource=lock.resource)


__all__ = [
    "RedisDistLockManager",
    "Lock",
]



from typing import List, Optional

import pytest
from pydantic import Field

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel, ProductModel
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_block_one(test_product_sub_block_one):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneForTest"):
        sub_block: SubBlockOneForTestInactive
        sub_block_2: Optional[SubBlockOneForTestInactive] = None
        sub_block_list: List[SubBlockOneForTestInactive] = []
        int_field: Optional[int] = None
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        sub_block: SubBlockOneForTestProvisioning
        sub_block_2: SubBlockOneForTestProvisioning
        sub_block_list: List[SubBlockOneForTestProvisioning]
        int_field: int
        str_field: Optional[str] = None
        list_field: List[int]

    class ProductBlockOneForTest(ProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        int_field: int
        str_field: str
        list_field: List[int]

    return ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest


@pytest.fixture
def test_product_block_one_db(resource_type_list, resource_type_int, resource_type_str, test_product_sub_block_one_db):
    product_block = ProductBlockTable(
        name="ProductBlockOneForTest", description="Test Block", tag="TEST", status="active"
    )

    product_block.resource_types = [resource_type_int, resource_type_str, resource_type_list]
    product_block.children = [test_product_sub_block_one_db]

    db.session.add(product_block)
    db.session.commit()

    return product_block, test_product_sub_block_one_db


@pytest.fixture
def test_product_model(test_product_one):
    return ProductModel(
        product_id=test_product_one,
        name="TestProductOne",
        description="Test ProductTable",
        product_type="Test",
        tag="TEST",
        status=ProductLifecycle.ACTIVE,
    )



from typing import TypeVar, Union

import pytest

from orchestrator.db import ProductTable, db
from orchestrator.db.models import FixedInputTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionInstanceList, SubscriptionModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_type_list_union(test_product_sub_block_one, test_product_sub_block_two):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    SubBlockTwoForTestInactive, SubBlockTwoForTestProvisioning, SubBlockTwoForTest = test_product_sub_block_two

    T = TypeVar("T", covariant=True)

    class ListOfPorts(SubscriptionInstanceList[T]):
        min_items = 1

    class ProductListUnionInactive(SubscriptionModel, is_base=True):
        test_fixed_input: bool
        list_union_blocks: ListOfPorts[Union[SubBlockTwoForTestInactive, SubBlockOneForTestInactive]]

    class ProductListUnionProvisioning(ProductListUnionInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_fixed_input: bool
        list_union_blocks: ListOfPorts[Union[SubBlockTwoForTestProvisioning, SubBlockOneForTestProvisioning]]

    class ProductListUnion(ProductListUnionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        list_union_blocks: ListOfPorts[Union[SubBlockTwoForTest, SubBlockOneForTest]]

    SUBSCRIPTION_MODEL_REGISTRY["ProductListUnion"] = ProductListUnion
    yield ProductListUnionInactive, ProductListUnionProvisioning, ProductListUnion
    del SUBSCRIPTION_MODEL_REGISTRY["ProductListUnion"]


@pytest.fixture
def test_product_list_union(test_product_block_one_db, test_product_sub_block_one_db, test_product_sub_block_two_db):
    product = ProductTable(
        name="ProductListUnion",
        description="Test List Union Product",
        product_type="Test",
        tag="Union",
        status="active",
    )

    fixed_input = FixedInputTable(name="test_fixed_input", value="False")

    product_block, _ = test_product_block_one_db
    product.fixed_inputs = [fixed_input]
    product.product_blocks = [product_block, test_product_sub_block_one_db, test_product_sub_block_two_db]
    db.session.add(product)
    db.session.commit()
    return product.product_id



import pathlib
import tempfile

from orchestrator.services.translations import generate_translations
from orchestrator.settings import app_settings


def test_generate_translations_empty():
    assert generate_translations("") == {}


def test_generate_translations():
    assert generate_translations("en-GB")


def test_generate_translations_custom_file():
    old_value = app_settings.TRANSLATIONS_DIR
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            with open(f"{temp_dir}/nl-NL.json", "w") as temp_file:
                temp_file.write("[]")

            app_settings.TRANSLATIONS_DIR = pathlib.Path(temp_dir)

            assert generate_translations("nl-NL") == {}

            with open(f"{temp_dir}/nl-NL.json", "w") as temp_file:
                temp_file.write('{"foo": "bar"}')

            assert generate_translations("nl-NL") == {"foo": "bar"}

            app_settings.TRANSLATIONS_DIR = old_value

    finally:
        app_settings.TRANSLATIONS_DIR = old_value



from datetime import datetime
from typing import List, Optional, TypeVar
from unittest import mock
from uuid import uuid4

import pytest
import pytz
from pydantic import Field, ValidationError, conlist
from pydantic.main import BaseModel
from pydantic.types import ConstrainedList
from sqlalchemy.orm.exc import NoResultFound

from orchestrator.db import (
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    db,
)
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import (
    ProductBlockModel,
    SubscriptionInstanceList,
    SubscriptionModel,
    _is_constrained_list_type,
)
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle


def test_product_block_metadata(test_product_block_one, test_product_one, test_product_block_one_db):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    subscription_id = uuid4()
    ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id
    )  # Need at least one instance since we lazy load this stuff

    product_block, _ = test_product_block_one_db

    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"


def test_lifecycle(test_product_model, test_product_type_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one
    subscription_id = uuid4()

    # Test create with wrong lifecycle, we can create
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestInactive(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL


def test_lifecycle_specific(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    _, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    _, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    # Works with less contrained lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL

    # Works with right lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.ACTIVE

    # Does not work with more constrained lifecycle
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestProvisioning(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestProvisioning.new(
                subscription_id=subscription_id,
                int_field=3,
                list_field=[1],
                sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
                sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            ),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestProvisioning.new(
            subscription_id=subscription_id,
            int_field=3,
            list_field=[1],
            sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )
    assert product_type.status == SubscriptionLifecycle.PROVISIONING


def test_product_blocks_per_lifecycle(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=3),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=uuid4(),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    with pytest.raises(
        ValidationError,
        match=r"2 validation errors for SubBlockOneForTest\nint_field\n  field required .+\nstr_field\n  field required .+",
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTest.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(ValidationError, match=r"5 validation errors for ProductTypeOneForTest"):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n  field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.INITIAL,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n  field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=uuid4(),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )


def test_change_lifecycle(test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product_type = ProductTypeOneForTestInactive.from_product_id(
        test_product_one,
        uuid4(),
    )
    product_type.block = ProductBlockOneForTestInactive.new(
        subscription_id=product_type.subscription_id,
        sub_block_2=SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id),
    )

    # Does not work if constraints are not met
    with pytest.raises(ValidationError, match=r"int_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    with pytest.raises(ValidationError, match=r"int_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)

    # Set first value
    product_type.block.int_field = 3
    product_type.block.sub_block.int_field = 3
    product_type.block.sub_block_2.int_field = 3
    product_type.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id)]
    product_type.block.sub_block_list[0].int_field = 4
    product_type.block.list_field = [1]

    # Does not work if constraints are not met
    with pytest.raises(ValidationError, match=r"str_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)

    # works with correct data
    product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)
    assert product_type.status == SubscriptionLifecycle.PROVISIONING
    assert product_type.block.str_field is None

    product_type.block.str_field = "A"
    product_type.block.sub_block.str_field = "B"
    product_type.block.sub_block_2.str_field = "C"
    product_type.block.sub_block_list[0].str_field = "D"

    # works with correct data
    product_type_new = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    assert product_type_new.status == SubscriptionLifecycle.ACTIVE
    expected_dict = product_type.dict()
    expected_dict["status"] = SubscriptionLifecycle.ACTIVE
    expected_dict["start_date"] = mock.ANY
    assert product_type_new.dict() == expected_dict
    assert isinstance(product_type_new.start_date, datetime)


def test_save_load(test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    customer_id = uuid4()

    model = ProductTypeOneForTestInactive.from_product_id(
        product_id=test_product_model.product_id,
        customer_id=customer_id,
        insync=True,
        description="Desc",
        start_date=datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        end_date=datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        note="Note",
        status=SubscriptionLifecycle.INITIAL,
    )

    # Set one value and make sure its saved
    model.block.str_field = "A"

    model.save()
    db.session.commit()

    assert (
        SubscriptionInstanceValueTable.query.join(SubscriptionInstanceValueTable.subscription_instance)
        .filter(SubscriptionInstanceTable.subscription_id == model.subscription_id)
        .count()
        == 1
    )

    assert model.dict() == {
        "block": {
            "int_field": None,
            "label": None,
            "list_field": [],
            "name": "ProductBlockOneForTest",
            "str_field": "A",
            "sub_block": {
                "int_field": None,
                "label": None,
                "name": "SubBlockOneForTest",
                "str_field": None,
                "subscription_instance_id": mock.ANY,
                "owner_subscription_id": mock.ANY,
            },
            "sub_block_2": None,
            "sub_block_list": [],
            "owner_subscription_id": mock.ANY,
            "subscription_instance_id": mock.ANY,
        },
        "customer_id": customer_id,
        "description": "Desc",
        "end_date": datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        "insync": True,
        "note": "Note",
        "product": {
            "description": "Test ProductTable",
            "name": "TestProductOne",
            "product_id": test_product_model.product_id,
            "product_type": "Test",
            "status": ProductLifecycle.ACTIVE,
            "tag": "TEST",
        },
        "start_date": datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        "status": SubscriptionLifecycle.INITIAL,
        "subscription_id": mock.ANY,
        "test_fixed_input": False,
    }

    # Set first value
    model.block.int_field = 3
    model.block.sub_block.int_field = 3
    model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=model.subscription_id)
    model.block.sub_block_2.int_field = 3
    model.block.list_field = [1]
    model.block.sub_block.str_field = "B"
    model.block.sub_block_2.str_field = "C"

    # works with correct data
    model = SubscriptionModel.from_other_lifecycle(model, SubscriptionLifecycle.ACTIVE)

    model.save()
    db.session.commit()

    new_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert model.dict() == new_model.dict()

    # Second save also works as expected
    new_model.save()
    db.session.commit()

    latest_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert new_model.dict() == latest_model.dict()

    # Loading blocks also works
    block = ProductBlockOneForTest.from_db(model.block.subscription_instance_id)
    assert block.dict() == model.block.dict()


def test_update_constrained_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestConListProductType(SubscriptionModel, is_base=True):
        saps: conlist(ProductBlockOneForTestInactive, min_items=1, max_items=4)

    # Creates
    ip = TestConListProductType.from_product_id(product_id=test_product_one, customer_id=uuid4())
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps, removes old one
    ip.saps = [sap]

    ip.save()

    ip2 = TestConListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()

    # Test constraint
    with pytest.raises(ValidationError):
        ip.saps = []


def test_update_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        saps: List[ProductBlockOneForTestInactive]

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=uuid4())
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps
    ip.saps = [sap]

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()


def test_update_optional(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        sap: Optional[ProductBlockOneForTestInactive] = None

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=uuid4())
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new sap
    ip.sap = sap

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()


def test_generic_from_subscription(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    ip = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=uuid4())
    ip.save()

    model = SubscriptionModel.from_subscription(ip.subscription_id)

    assert isinstance(model, ProductTypeOneForTestInactive)


def test_label_is_saved(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(test_product_one, uuid4())
    test_model.block.label = "My label"
    test_model.save()
    db.session.commit()
    instance_in_db = SubscriptionInstanceTable.query.get(test_model.block.subscription_instance_id)
    assert instance_in_db.label == "My label"


def test_domain_model_attrs_saving_loading(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=uuid4())
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    test_model.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)]
    test_model.save()
    db.session.commit()

    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.child_id == test_model.block.sub_block.subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block"
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.child_id == test_model.block.sub_block_2.subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block_2"
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.child_id == test_model.block.sub_block_list[0].subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block_list"
    test_model_2 = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert test_model == test_model_2


def test_removal_of_domain_attrs(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=uuid4())
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    test_model.save()
    db.session.commit()
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.child_id == test_model.block.sub_block.subscription_instance_id
    ).one()
    relation.domain_model_attr = None
    db.session.commit()
    with pytest.raises(ValueError, match=r"Expected exactly one item in iterable, but got"):
        ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)


def test_simple_model_with_no_attrs(generic_subscription_1, generic_product_type_1):
    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    model = GenericProductOne.from_subscription(subscription_id=generic_subscription_1)
    with pytest.raises(NoResultFound):
        SubscriptionInstanceRelationTable.query.filter(
            SubscriptionInstanceRelationTable.child_id == model.pb_1.subscription_instance_id
        ).one()


def test_abstract_super_block(test_product_one, test_product_type_one, test_product_block_one_db):
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    class AbstractProductBlockOneForTestInactive(ProductBlockModel):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: Optional[str] = None
        list_field: List[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)
        int_field: Optional[int] = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: Optional[str] = None
        list_field: List[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]
        int_field: int

    class AbstractProductTypeOneForTestInactive(SubscriptionModel):
        block: AbstractProductBlockOneForTestInactive

    class AbstractProductTypeOneForTestProvisioning(AbstractProductTypeOneForTestInactive):
        block: AbstractProductBlockOneForTestProvisioning

    class AbstractProductTypeOneForTest(AbstractProductTypeOneForTestProvisioning):
        block: AbstractProductBlockOneForTest

    class ProductTypeOneForTestInactive(AbstractProductTypeOneForTestInactive, is_base=True):
        block: ProductBlockOneForTestInactive

    class ProductTypeOneForTestProvisioning(
        ProductTypeOneForTestInactive,
        AbstractProductTypeOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        block: ProductBlockOneForTestProvisioning

    class ProductTypeOneForTest(
        ProductTypeOneForTestProvisioning, AbstractProductTypeOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=uuid4())
    test_model.block = ProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    product_block, product_sub_block = test_product_block_one_db

    # Check metadata
    with pytest.raises(
        ValueError, match=r"Cannot create instance of abstract class. Use one of {'ProductBlockOneForTest'}"
    ):
        AbstractProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    assert AbstractProductBlockOneForTestInactive.name is None
    assert not hasattr(AbstractProductBlockOneForTestInactive, "description")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "product_block_id")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "tag")
    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"

    test_model.save()
    db.session.commit()

    test_model = AbstractProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model, ProductTypeOneForTestInactive)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model.block.int_field = 1
    test_model.block.str_field = "bla"
    test_model.block.list_field = [1]

    test_model = SubscriptionModel.from_other_lifecycle(test_model, SubscriptionLifecycle.ACTIVE)
    test_model.save()
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = AbstractProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = ProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    block = AbstractProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.dict() == test_model.block.dict()
    assert isinstance(block, ProductBlockOneForTest)

    block = ProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.dict() == test_model.block.dict()
    assert isinstance(block, ProductBlockOneForTest)


def test_subscription_instance_list():
    T = TypeVar("T")

    class Max2List(SubscriptionInstanceList[T]):
        max_items = 2

    # Also check generic subclass
    class ConList(Max2List[T]):
        min_items = 1

    class Model(BaseModel):
        list_field: ConList[int]

    with pytest.raises(ValidationError):
        Model(list_field=["a"])

    Model(list_field=[1])


def test_is_constrained_list_type():
    class ListType(ConstrainedList):
        min_items = 1

    assert _is_constrained_list_type(ListType) is True
    assert _is_constrained_list_type(SubscriptionInstanceList[int]) is True
    assert _is_constrained_list_type(Optional[int]) is False
    assert _is_constrained_list_type(List[int]) is False


# TODO #1320: Unskip test after updating diff_product_in_database function
@pytest.mark.skip
def test_diff_in_db(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    assert ProductTypeOneForTestInactive.diff_product_in_database(test_product_one) == {}

    class Wrong(SubscriptionModel):
        pass

    assert (
        Wrong.diff_product_in_database(test_product_one)
        == {
            "TestProductOne": {
                "missing_fixed_inputs_in_model": {"test_fixed_input"},
                "missing_product_blocks_in_model": {"ProductBlockOneForTest"},
            }
        }
        != {
            "TestProductOne": {
                "missing_in_children": {
                    "ProductBlockOneForTest": {"missing_product_blocks_in_db": {"SubBlockOneForTest"}}
                },
                "missing_product_blocks_in_model": {"SubBlockOneForTest"},
            }
        }
    )


# TODO #1320: Unskip test after updating diff_product_in_database function
@pytest.mark.skip
def test_diff_in_db_missing_in_db(test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product = ProductTable(
        name="TestProductEmpty", description="Test ProductTable", product_type="Test", tag="TEST", status="active"
    )

    db.session.add(product)
    db.session.commit()

    assert ProductTypeOneForTestInactive.diff_product_in_database(product.product_id) == {
        "TestProductEmpty": {
            "missing_fixed_inputs_in_db": {"test_fixed_input"},
            "missing_in_children": {
                "ProductBlockOneForTest": {
                    "missing_product_blocks_in_db": {"SubBlockOneForTest"},
                    "missing_resource_types_in_db": {"int_field", "list_field", "str_field"},
                },
                "SubBlockOneForTest": {"missing_resource_types_in_db": {"int_field", "str_field"}},
            },
            "missing_product_blocks_in_db": {"ProductBlockOneForTest"},
        }
    }


def test_from_other_lifecycle_abstract(test_product_one):
    class AbstractProductBlockOneForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneForTest"):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: Optional[str] = None
        list_field: List[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)
        int_field: Optional[int] = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: Optional[str] = None
        list_field: List[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]
        int_field: int

    block = ProductBlockOneForTestInactive.new(subscription_id=uuid4())
    assert isinstance(block, ProductBlockOneForTestInactive)

    block.int_field = 1
    block.str_field = "bla"
    block.list_field = [1]

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, AbstractProductBlockOneForTest)
    assert isinstance(active_block, ProductBlockOneForTest)
    assert active_block.db_model == block.db_model


def test_from_other_lifecycle_sub(test_product_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    subscription_id = uuid4()

    block = ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id, int_field=1, str_field="bla", list_field=[1]
    )
    block.sub_block = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_list = [
        SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    ]

    assert isinstance(block, ProductBlockOneForTestInactive)

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, ProductBlockOneForTest)
    assert isinstance(active_block.sub_block, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_2, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_list[0], SubBlockOneForTest)
    assert active_block.db_model == block.db_model
    assert active_block.sub_block.db_model == block.sub_block.db_model
    assert active_block.sub_block_2.db_model == block.sub_block_2.db_model
    assert active_block.sub_block_list[0].db_model == block.sub_block_list[0].db_model



from typing import Optional

import pytest

from orchestrator.db import ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_type_sub_list_union(test_product_block_with_list_union):
    (
        ProductBlockWithListUnionForTestInactive,
        ProductBlockWithListUnionForTestProvisioning,
        ProductBlockWithListUnionForTest,
    ) = test_product_block_with_list_union

    class ProductSubListUnionInactive(SubscriptionModel, is_base=True):
        test_block: Optional[ProductBlockWithListUnionForTestInactive]

    class ProductSubListUnionProvisioning(ProductSubListUnionInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_block: ProductBlockWithListUnionForTestProvisioning

    class ProductSubListUnion(ProductSubListUnionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockWithListUnionForTest

    SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"] = ProductSubListUnion
    yield ProductSubListUnionInactive, ProductSubListUnionProvisioning, ProductSubListUnion
    del SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"]


@pytest.fixture
def test_product_sub_list_union(test_product_block_with_list_union_db):
    product = ProductTable(
        name="ProductSubListUnion",
        description="Product with Union sub product_block",
        tag="UnionSub",
        product_type="Test",
        status="active",
    )
    product_block, _, _ = test_product_block_with_list_union_db
    product.product_blocks = [product_block]
    db.session.add(product)
    db.session.commit()

    return product.product_id



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import secrets
import string
from pathlib import Path
from typing import List, Optional

from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from pydantic import BaseSettings


class AppSettings(BaseSettings):
    TESTING: bool = True
    SESSION_SECRET: str = "".join(secrets.choice(string.ascii_letters) for i in range(16))  # noqa: S311
    CORS_ORIGINS: str = "*"
    CORS_ALLOW_METHODS: List[str] = ["GET", "PUT", "PATCH", "POST", "DELETE", "OPTIONS", "HEAD"]
    CORS_ALLOW_HEADERS: List[str] = ["If-None-Match", "Authorization", "If-Match", "Content-Type"]
    CORS_EXPOSE_HEADERS: List[str] = [
        "Cache-Control",
        "Content-Language",
        "Content-Length",
        "Content-Type",
        "Expires",
        "Last-Modified",
        "Pragma",
        "Content-Range",
        "ETag",
    ]
    ENVIRONMENT: str = "local"
    WORKFLOWS_SWAGGER_HOST: str = "localhost"
    WORKFLOWS_GUI_URI: str = "http://localhost:3000"
    DATABASE_URI: str = "postgresql://nwa:nwa@localhost/orchestrator-core"
    MAX_WORKERS: int = 5
    MAIL_SERVER: str = "localhost"
    MAIL_PORT: int = 25
    MAIL_STARTTLS: bool = False
    CACHE_HOST: str = "127.0.0.1"
    CACHE_PORT: int = 6379
    ENABLE_DISTLOCK_MANAGER: bool = True
    DISTLOCK_BACKEND: str = "memory"
    CC_NOC: int = 0
    SERVICE_NAME: str = "orchestrator-core"
    LOGGING_HOST: str = "localhost"
    LOG_LEVEL: str = "DEBUG"
    SLACK_ENGINE_SETTINGS_HOOK_ENABLED: bool = False
    SLACK_ENGINE_SETTINGS_HOOK_URL: str = ""
    TRACING_ENABLED: bool = False
    TRANSLATIONS_DIR: Optional[Path] = None
    WEBSOCKET_BROADCASTER_URL: str = "memory://"
    ENABLE_WEBSOCKETS: bool = True
    DISABLE_INSYNC_CHECK: bool = False


class Oauth2Settings(BaseSettings):
    OAUTH2_ACTIVE: bool = False
    OAUTH2_RESOURCE_SERVER_ID: str = ""
    OAUTH2_RESOURCE_SERVER_SECRET: str = ""
    OAUTH2_TOKEN_URL: str = ""
    OIDC_CONF_WELL_KNOWN_URL: str = ""
    OPA_URL: str = "http://127.0.0.1:8181/v1/data/automation/authorization/allow"


app_settings = AppSettings()
oauth2_settings = Oauth2Settings()

# Tracer settings
tracer_provider = TracerProvider()

jaeger_exporter = JaegerExporter(agent_host_name=app_settings.LOGGING_HOST, udp_split_oversized_batches=True)
tracer_provider.add_span_processor(BatchSpanProcessor(jaeger_exporter))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import List
from uuid import UUID

from fastapi.param_functions import Body
from fastapi.routing import APIRouter

from orchestrator.api.error_handling import raise_status
from orchestrator.api.models import delete, save, update
from orchestrator.db import ProductBlockTable, ResourceTypeTable
from orchestrator.schemas import ResourceTypeBaseSchema, ResourceTypeSchema

router = APIRouter()


@router.get("/", response_model=List[ResourceTypeSchema])
def fetch() -> List[ResourceTypeTable]:
    return ResourceTypeTable.query.all()


@router.get("/{resource_type_id}", response_model=ResourceTypeSchema)
def resource_type_by_id(resource_type_id: UUID) -> ResourceTypeTable:
    resource_type = ResourceTypeTable.query.filter_by(resource_type_id=resource_type_id).first()
    if not resource_type:
        raise_status(HTTPStatus.NOT_FOUND, f"Resource type {resource_type_id} not found")
    return resource_type


@router.post("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def save_resource_type(data: ResourceTypeBaseSchema = Body(...)) -> None:
    return save(ResourceTypeTable, data)


@router.put("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def update_resource_type(data: ResourceTypeSchema = Body(...)) -> None:
    return update(ResourceTypeTable, data)


@router.delete("/{resource_type_id}", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def delete_resource_type(resource_type_id: UUID) -> None:
    product_blocks = ProductBlockTable.query.filter(
        ProductBlockTable.resource_types.any(ResourceTypeTable.resource_type_id == str(resource_type_id))
    ).all()
    if len(product_blocks) > 0:
        error_products_block = ", ".join(map(lambda pb: pb.name, product_blocks))
        raise_status(
            HTTPStatus.BAD_REQUEST, f"ResourceType {resource_type_id} is used in ProductBlocks: {error_products_block}"
        )
    return delete(ResourceTypeTable, resource_type_id)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from datetime import datetime

import pytz

TIMESTAMP_REGEX = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}")


def isoformat(dt: datetime) -> str:
    """ISO format datetime object with max precision limited to seconds.

    Args:
        dt: datatime object to be formatted

    Returns:
        ISO 8601 formatted string

    """
    # IMPORTANT should the format be ever changed, be sure to update TIMESTAMP_REGEX as well!
    return dt.isoformat(timespec="seconds")


def timestamp() -> str:
    """Return iso formatted UTC timestamp as a string.

    Note: precision is in seconds.

    >>> assert TIMESTAMP_REGEX.match(timestamp())

    Returns:
        iso formatted UTC timestamp.

    """
    return isoformat(datetime.utcnow())


def nowtz() -> datetime:
    """Fetch Datetime now object in UTC.

    Returns:
        Datetime object

    """
    return datetime.now(tz=pytz.utc)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from time import sleep

import schedule
import typer

from orchestrator.schedules import ALL_SCHEDULERS

log = logging.getLogger(__name__)

app: typer.Typer = typer.Typer()


@app.command()
def run() -> None:
    """Loop eternally and run schedulers at configured times."""
    for s in ALL_SCHEDULERS:
        job = getattr(schedule.every(s.period), s.time_unit)
        if s.at:
            job = job.at(s.at)
        job.do(s).tag(s.name)
    log.info("Starting Schedule")
    for j in schedule.jobs:
        log.info("%s: %s", ", ".join(j.tags), j)
    while True:
        schedule.run_pending()
        idle = schedule.idle_seconds()
        if idle < 0:
            log.info("Next job in queue is scheduled in the past, run it now.")
        else:
            log.info("Sleeping for %d seconds", idle)
            sleep(idle)


@app.command()
def show_schedule() -> None:
    """Show the currently configured schedule."""
    for s in ALL_SCHEDULERS:
        at_str = f"@ {s.at} " if s.at else ""
        typer.echo(f"{s.name}: {s.__name__} {at_str}every {s.period} {s.time_unit}")


@app.command()
def force(keyword: str) -> None:
    """Force the execution of (a) scheduler(s) based on a keyword."""
    for s in ALL_SCHEDULERS:
        if keyword in s.name or keyword in s.__name__:
            s()



import time
import uuid
from http import HTTPStatus
from threading import Condition, Event
from unittest import mock
from uuid import uuid4

import pytest

from orchestrator.config.assignee import Assignee
from orchestrator.db import (
    EngineSettingsTable,
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    SubscriptionTable,
    WorkflowTable,
    db,
)
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus, done, init, step, workflow
from test.unit_tests.conftest import CUSTOMER_ID
from test.unit_tests.workflows import WorkflowInstanceForTests

test_condition = Condition()


@pytest.fixture
def long_running_workflow():
    @step("Long Running Step")
    def long_running_step():
        with test_condition:
            test_condition.wait()
        return {"done": True}

    @workflow("Long Running Workflow")
    def long_running_workflow_py():
        return init >> long_running_step >> long_running_step >> done

    with WorkflowInstanceForTests(long_running_workflow_py, "long_running_workflow_py"):

        db_workflow = WorkflowTable(name="long_running_workflow_py", target=Target.MODIFY)
        db.session.add(db_workflow)
        db.session.commit()

        yield "long_running_workflow_py"


@pytest.fixture
def started_process(test_workflow, generic_subscription_1):
    pid = uuid4()
    process = ProcessTable(pid=pid, workflow=test_workflow, last_status=ProcessStatus.SUSPENDED)
    init_step = ProcessStepTable(pid=pid, name="Start", status="success", state={})
    insert_step = ProcessStepTable(
        pid=pid, name="Insert UUID in state", status="success", state={"subscription_id": generic_subscription_1}
    )
    check_step = ProcessStepTable(
        pid=pid,
        name="Test that it is a string now",
        status="success",
        state={"subscription_id": generic_subscription_1},
    )
    step = ProcessStepTable(pid=pid, name="Modify", status="suspend", state={"subscription_id": generic_subscription_1})

    process_subscription = ProcessSubscriptionTable(pid=pid, subscription_id=generic_subscription_1)

    db.session.add(process)
    db.session.add(init_step)
    db.session.add(insert_step)
    db.session.add(check_step)
    db.session.add(step)
    db.session.add(process_subscription)
    db.session.commit()

    return pid


def test_show(test_client, started_process):
    response = test_client.get(f"/api/processes/{started_process}")
    assert HTTPStatus.OK == response.status_code
    process = response.json()
    assert "workflow_for_testing_processes_py", process["workflow_name"]


def test_show_invalid_uuid(test_client):
    response = test_client.get("/api/processes/wrong")
    assert response.status_code == 422


def test_show_not_found(test_client, started_process):
    response = test_client.get("/api/processes/120d31e4-9166-47cb-ad37-b78072c7ab8b")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_delete_process(responses, test_client, started_process):
    processes = test_client.get("/api/processes").json()
    before_delete_count = len(processes)

    response = test_client.delete(f"/api/processes/{started_process}")
    assert HTTPStatus.NO_CONTENT == response.status_code
    assert before_delete_count - 1 == len(test_client.get("/api/processes").json())


def test_delete_process_404(test_client, started_process):
    response = test_client.delete(f"/api/processes/{uuid4()}")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_long_running_pause(test_client, long_running_workflow):
    app_settings.TESTING = False
    # Start the workflow
    response = test_client.post(f"/api/processes/{long_running_workflow}", json=[{}])
    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    pid = response.json()["id"]

    response = test_client.get(f"api/processes/{pid}")
    assert HTTPStatus.OK == response.status_code

    # Let it run untill the first lock step is started
    time.sleep(1)

    response = test_client.put("/api/settings/status", json={"global_lock": True})
    assert response.json()["global_lock"] is True
    assert response.json()["running_processes"] == 1
    assert response.json()["global_status"] == "PAUSING"

    # Let it run until completing the first lock step
    with test_condition:
        test_condition.notify_all()
    time.sleep(1)

    # Make sure the running_processes is decreased and status is updated.
    response = test_client.get("/api/settings/status")
    assert response.json()["global_lock"] is True
    assert response.json()["running_processes"] == 0
    assert response.json()["global_status"] == "PAUSED"

    response = test_client.get(f"api/processes/{pid}")
    assert len(response.json()["steps"]) == 4
    assert response.json()["current_state"]["done"] is True
    # assume ordered steplist
    assert response.json()["steps"][3]["status"] == "pending"

    response = test_client.put("/api/settings/status", json={"global_lock": False})

    # Make sure it started again
    time.sleep(1)

    assert response.json()["global_lock"] is False
    assert response.json()["running_processes"] == 1
    assert response.json()["global_status"] == "RUNNING"

    # Let it finish after second lock step
    with test_condition:
        test_condition.notify_all()
    time.sleep(1)

    response = test_client.get(f"api/processes/{pid}")
    assert HTTPStatus.OK == response.status_code
    # assume ordered steplist
    assert response.json()["steps"][3]["status"] == "complete"

    app_settings.TESTING = True


def test_service_unavailable_engine_locked(test_client, test_workflow):
    engine_settings = EngineSettingsTable.query.one()
    engine_settings.global_lock = True
    db.session.flush()

    response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

    assert HTTPStatus.SERVICE_UNAVAILABLE == response.status_code


def test_complete_workflow(test_client, test_workflow):
    response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    pid = response.json()["id"]

    response = test_client.get(f"api/processes/{pid}")
    assert HTTPStatus.OK == response.status_code

    process = response.json()
    assert Assignee.CHANGES == process["assignee"]
    assert "suspended" == process["status"]

    steps = process["steps"]
    assert "success" == steps[0]["status"]

    # Check validation
    user_input = {
        "generic_select": 123,
    }

    response = test_client.put(f"/api/processes/{pid}/resume", json=[user_input])
    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert response.json()["validation_errors"] == [
        {
            "loc": ["generic_select"],
            "msg": "value is not a valid enumeration member; permitted: 'A', 'B', 'C'",
            "type": "type_error.enum",
            "ctx": {"enum_values": ["A", "B", "C"]},
        }
    ]

    response = test_client.get(f"api/processes/{pid}")

    process = response.json()
    assert "suspended" == process["status"]

    # Now for real
    user_input = {"generic_select": "A"}

    response = test_client.put(f"/api/processes/{pid}/resume", json=[user_input])
    assert HTTPStatus.NO_CONTENT == response.status_code

    process = test_client.get(f"api/processes/{pid}").json()
    assert "completed" == process["status"]


def test_abort_process(test_client, started_process):
    response = test_client.put(f"/api/processes/{started_process}/abort")
    assert HTTPStatus.NO_CONTENT == response.status_code

    aborted_process = test_client.get(f"/api/processes/{started_process}").json()
    assert "aborted" == aborted_process["status"]


def test_process_subscription_by_subscription_id(test_client, started_process, generic_subscription_1):
    response = test_client.get(f"/api/processes/process-subscriptions-by-subscription-id/{generic_subscription_1}")
    assert HTTPStatus.OK == response.status_code
    process_subscriptions = response.json()
    assert 1 == len(process_subscriptions)
    assert process_subscriptions[0]["subscription_id"].lower(), generic_subscription_1
    assert process_subscriptions[0]["pid"].lower(), started_process
    assert process_subscriptions[0]["workflow_target"], Target.CREATE
    assert process_subscriptions[0]["process"]["workflow"], "workflow_for_testing_processes_py"


def test_process_subscription_by_pid(test_client, started_process, generic_subscription_1):
    response = test_client.get(f"/api/processes/process-subscriptions-by-pid/{started_process}")
    assert HTTPStatus.OK == response.status_code
    process_subscriptions = response.json()
    assert 1 == len(process_subscriptions)
    assert process_subscriptions[0]["subscription_id"].lower(), generic_subscription_1
    assert process_subscriptions[0]["workflow_target"], Target.CREATE


def test_process_subscription_by_pid_404(test_client):
    response = test_client.get(f"/api/processes/process-subscriptions-by-pid/{uuid4()}")
    assert 0 == len(response.json())


def test_process_subscription_by_subscription_id_404(test_client):
    response = test_client.get(f"/api/processes/process-subscriptions-by-subscription-id/{uuid4()}")
    assert 0 == len(response.json())


def test_new_process_invalid_body(test_client, long_running_workflow):
    response = test_client.post(f"/api/processes/{long_running_workflow}", json=[{"wrong": "body"}])
    assert HTTPStatus.BAD_REQUEST == response.status_code


def test_new_process_invalid_Content_Type(test_client):
    response = test_client.post("/api/processes/terminate_sn8_light_path")
    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code


def test_new_process_post_inconsistent_data(test_client):
    response = test_client.post("/api/processes/terminate_sn8_light_path", json={})
    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code


def test_404_resume(test_client):
    response = test_client.put(f"/api/processes/{uuid.uuid4()}/resume", json={})
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_resume_validations(test_client, started_process):
    process_info_before = test_client.get(f"/api/processes/{started_process}").json()

    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": 123}])
    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert [
        {
            "ctx": {"enum_values": ["A", "B", "C"]},
            "loc": ["generic_select"],
            "msg": "value is not a valid enumeration member; permitted: 'A', 'B', 'C'",
            "type": "type_error.enum",
        }
    ] == response.json()["validation_errors"]
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) == len(excuted_steps_before)
    assert process_info_after["status"] == "suspended"


def test_resume_with_empty_form(test_client, started_process):
    # Set a default value for the only input so we can submit an empty form
    step = ProcessStepTable.query.filter(
        ProcessStepTable.name == "Modify", ProcessStepTable.pid == started_process
    ).one()
    step.state["generic_select"] = "A"
    db.session.add(step)
    db.session.flush()

    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])

    assert HTTPStatus.NO_CONTENT == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) > len(excuted_steps_before)
    assert process_info_after["status"] == "completed"


def test_resume_happy_flow(test_client, started_process):
    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])

    assert HTTPStatus.NO_CONTENT == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) > len(excuted_steps_before)
    assert process_info_after["status"] == "completed"


def test_resume_with_incorrect_workflow_status(test_client, started_process):
    process = ProcessTable.query.get(started_process)
    # setup DB so it looks like this workflow is already resumed
    process.last_status = ProcessStatus.RUNNING
    process.failed_reason = ""
    db.session.commit()

    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])
    assert 409 == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) == len(excuted_steps_before)
    assert process_info_after["status"] == "running"


def test_processes_filterable(test_client, mocked_processes, generic_subscription_2, generic_subscription_1):
    response = test_client.get("/api/processes/")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()

    assert 7 == len(processes)
    assert "workflow_for_testing_processes_py", processes[0]["workflow"]

    response = test_client.get("/api/processes?filter=status,completed")
    assert 3 == len(response.json())
    response = test_client.get("/api/processes?filter=status,suspended")
    assert 2 == len(response.json())

    product_name = SubscriptionTable.query.get(generic_subscription_1).product.name
    response = test_client.get(f"/api/processes?filter=product,{product_name}")
    assert 3 == len(response.json())
    response = test_client.get("/api/processes?sort=assignee,asc")
    assert response.json()[0]["assignee"] == "NOC"
    response = test_client.get("/api/processes?sort=started&filter=istask,y")
    assert 3 == len(response.json())
    response = test_client.get(f"/api/processes?filter=istask,y,organisation,{CUSTOMER_ID}")
    assert 2 == len(response.json())


def test_processes_filterable_response_model(
    test_client, mocked_processes, generic_subscription_2, generic_subscription_1
):
    response = test_client.get("/api/processes/?sort=started,asc")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()
    assert len(processes) == 7
    process = processes[0]

    assert len(process["subscriptions"]) == 1

    # Check if the other fields are filled with correct data
    del process["pid"]  # skip pid as it's dynamic
    del process["subscriptions"]
    assert process == {
        "assignee": "SYSTEM",
        "created_by": None,
        "failed_reason": None,
        "last_modified_at": mock.ANY,
        "started_at": mock.ANY,
        "last_status": "completed",
        "last_step": "Modify",
        "workflow": "workflow_for_testing_processes_py",
        "workflow_target": "CREATE",
        "is_task": False,
    }


def test_processes_filterable_response_model_contains_product_info(
    test_client, mocked_processes, generic_subscription_2, generic_subscription_1
):
    response = test_client.get("/api/processes/?sort=started,asc")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()
    assert len(processes) == 7
    process = processes[0]

    assert len(process["subscriptions"]) == 1
    assert process["subscriptions"][0]["product"]["tag"] == "GEN1"
    assert process["subscriptions"][0]["product"]["name"] == "Product 1"


def test_processes_assignees(test_client):
    response = test_client.get("/api/processes/assignees")
    assert HTTPStatus.OK == response.status_code


def test_processes_statuses(test_client):
    response = test_client.get("/api/processes/statuses")
    assert HTTPStatus.OK == response.status_code


def test_resume_all_processes(test_client, mocked_processes_resumeall):
    """Test resuming all processes."""
    response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 3


def test_resume_all_processes_multiple_calls(test_client, mocked_processes_resumeall):
    """Test only 1 of multiple resume-all calls is successful.

    This uses the "MemoryDistlockManager" reference implementation.
    """
    event = Event()

    def resume_noop(*args, **kwargs):
        event.wait(2)  # To keep the lock open for a while

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False

    with mock.patch("orchestrator.services.processes.resume_process", new=resume_noop):
        responses = [test_client.put("/api/processes/resume-all") for _ in range(5)]
        responses.sort(key=lambda r: r.status_code)
        event.set()

    app_settings.TESTING = True

    assert responses[0].status_code == HTTPStatus.OK
    assert responses[0].json()["count"] == 3
    assert all(response.status_code == HTTPStatus.CONFLICT for response in responses[1:])


def test_resume_all_processes_nothing_to_do(test_client):
    """Test resuming all process when there are no process to be resumed."""
    response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 0


def test_resume_all_processes_value_error(test_client, mocked_processes_resumeall, caplog):
    """Test resuming all processes where one raises ValueError."""
    with mock.patch("orchestrator.services.processes.resume_process") as mocked_resume:
        mocked_resume.side_effect = [None, ValueError("This workflow cannot be resumed"), None]
        response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 3  # returns 3 because it's async
    assert "Failed to resume process" in caplog.text  # log should confirm 1 process was not resumed
    assert "Completed resuming processes" in caplog.text



from uuid import uuid4

import pytest
from sqlalchemy.orm.exc import MultipleResultsFound

from orchestrator.db import ProductTable
from orchestrator.services.subscriptions import get_subscription, retrieve_subscription_by_subscription_instance_value
from test.unit_tests import fixtures

CORRECT_SUBSCRIPTION = str(uuid4())
INCORRECT_SUBSCRIPTION = str(uuid4())

subscription_mapping = {"PB_2": [{"rt_3": "info.id", "rt_2": "info2.id"}]}


def test_get_subscription_by_id(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    subscription = fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )
    result = get_subscription(str(subscription.subscription_id))
    assert str(result.subscription_id) == CORRECT_SUBSCRIPTION


def test_get_subscription_by_id_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    with pytest.raises(ValueError):
        get_subscription(INCORRECT_SUBSCRIPTION)


def test_retrieve_subscription_by_subscription_instance_value_none(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    assert retrieve_subscription_by_subscription_instance_value("rt_2", "Wrong") is None


def test_retrieve_subscription_by_subscription_instance_value(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    assert (
        str(retrieve_subscription_by_subscription_instance_value("rt_3", "0").subscription_id) == CORRECT_SUBSCRIPTION
    )


def test_retrieve_subscription_by_subscription_instance_value_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    with pytest.raises(MultipleResultsFound):
        retrieve_subscription_by_subscription_instance_value("rt_3", "0")



#!/usr/bin/env python3

# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Optional, Type

import sentry_sdk
import structlog
import typer
from fastapi.applications import FastAPI
from fastapi_etag.dependency import add_exception_handler
from nwastdlib.logging import initialise_logging
from opentelemetry import trace
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor
from opentelemetry.instrumentation.redis import RedisInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.responses import JSONResponse, Response

from orchestrator.api.api_v1.api import api_router
from orchestrator.api.error_handling import ProblemDetailException
from orchestrator.cli.main import app as cli_app
from orchestrator.db import db, init_database
from orchestrator.db.database import DBSessionMiddleware
from orchestrator.distlock import init_distlock_manager
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY, SubscriptionModel
from orchestrator.exception_handlers import form_error_handler, problem_detail_handler
from orchestrator.forms import FormException
from orchestrator.settings import AppSettings, app_settings, tracer_provider
from orchestrator.version import GIT_COMMIT_HASH
from orchestrator.websocket import init_websocket_manager

logger = structlog.get_logger(__name__)


class OrchestratorCore(FastAPI):
    def __init__(
        self,
        title: str = "The Orchestrator",
        description: str = "The orchestrator is a project that enables users to run workflows.",
        openapi_url: str = "/api/openapi.json",
        docs_url: str = "/api/docs",
        redoc_url: str = "/api/redoc",
        version: str = "1.0.0",
        default_response_class: Type[Response] = JSONResponse,
        base_settings: AppSettings = app_settings,
        **kwargs: Any,
    ) -> None:
        websocket_manager = init_websocket_manager(base_settings)
        distlock_manager = init_distlock_manager(base_settings)
        super().__init__(
            title=title,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            redoc_url=redoc_url,
            version=version,
            default_response_class=default_response_class,
            on_startup=[websocket_manager.connect_redis, distlock_manager.connect_redis],
            on_shutdown=[
                websocket_manager.disconnect_redis,
                websocket_manager.disconnect_all,
                distlock_manager.disconnect_redis,
            ],
            **kwargs,
        )

        initialise_logging()

        self.include_router(api_router, prefix="/api")

        init_database(base_settings)

        self.add_middleware(SessionMiddleware, secret_key=base_settings.SESSION_SECRET)
        self.add_middleware(DBSessionMiddleware, database=db)
        origins = base_settings.CORS_ORIGINS.split(",")
        self.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_methods=base_settings.CORS_ALLOW_METHODS,
            allow_headers=base_settings.CORS_ALLOW_HEADERS,
            expose_headers=base_settings.CORS_EXPOSE_HEADERS,
        )

        self.add_exception_handler(FormException, form_error_handler)
        self.add_exception_handler(ProblemDetailException, problem_detail_handler)
        add_exception_handler(self)

        @self.router.get("/", response_model=str, response_class=JSONResponse, include_in_schema=False)
        def _index() -> str:
            return "Orchestrator orchestrator"

    def instrument_app(self) -> None:
        logger.info("Activating Opentelemetry tracing to app", app=self.title)
        trace.set_tracer_provider(tracer_provider)
        FastAPIInstrumentor.instrument_app(self)
        RequestsInstrumentor().instrument()
        HTTPXClientInstrumentor().instrument()
        RedisInstrumentor().instrument()
        Psycopg2Instrumentor().instrument()
        SQLAlchemyInstrumentor().instrument(engine=db.engine, tracer_provider=tracer_provider)

    def add_sentry(
        self,
        sentry_dsn: str,
        trace_sample_rate: float,
        server_name: str,
        environment: str,
        release: Optional[str] = GIT_COMMIT_HASH,
    ) -> None:
        logger.info("Adding Sentry middleware to app", app=self.title)
        sentry_sdk.init(
            dsn=sentry_dsn,
            traces_sample_rate=trace_sample_rate,
            server_name=server_name,
            environment=environment,
            release=f"orchestrator@{release}",
            integrations=[SqlalchemyIntegration(), RedisIntegration()],
        )
        self.add_middleware(SentryAsgiMiddleware)

    @staticmethod
    def register_subscription_models(product_to_subscription_model_mapping: Dict[str, Type[SubscriptionModel]]) -> None:
        """
        Register your subscription models.

        This method is needed to register your subscription models inside the orchestrator core.

        Args:
            product_to_subscription_model_mapping: The dictionary should contain a mapping of products to SubscriptionModels.
                The selection will be done depending on the name of the product.

        Returns:
            None:

        Examples:
            >>> product_to_subscription_model_mapping = { # doctest:+SKIP
            ...     "Generic Product One": GenericProductModel,
            ...     "Generic Product Two": GenericProductModel,
            ... }

        """
        SUBSCRIPTION_MODEL_REGISTRY.update(product_to_subscription_model_mapping)


main_typer_app = typer.Typer()
main_typer_app.add_typer(cli_app, name="orchestrator", help="The orchestrator CLI commands")

if __name__ == "__main__":
    main_typer_app()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that provides service functions on subscriptions."""

from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Tuple, TypeVar, Union, overload
from uuid import UUID

import more_itertools
import structlog
from sqlalchemy import Text, cast, not_
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Query, aliased, joinedload
from sqlalchemy.sql.expression import or_

from orchestrator.db import (
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import SubscriptionInstanceRelationTable
from orchestrator.targets import Target
from orchestrator.types import UUIDstr
from orchestrator.utils.datetime import nowtz

logger = structlog.get_logger(__name__)


T = TypeVar("T", bound=SubscriptionTable)


@overload
def get_subscription(subscription_id: Union[UUID, UUIDstr], for_update: bool = False) -> SubscriptionTable:
    ...


@overload
def get_subscription(
    subscription_id: Union[UUID, UUIDstr], for_update: bool = False, model: T = SubscriptionTable
) -> T:
    ...


def get_subscription(
    subscription_id: Union[UUID, UUIDstr], for_update: bool = False, model: T = SubscriptionTable
) -> T:
    """Get the subscription.

    Args:
        subscription_id: The subscription_id
        for_update: specific whether we intend to update the subscription
        model: SubscriptionModelType

    Returns: A subscription object

    Raises: ValueError: if the requested Subscription does not exist in de database.

    """
    query = model.query
    if for_update:
        query = query.with_for_update()

    try:
        subscription = query.get(subscription_id)
    except SQLAlchemyError as e:
        raise ValueError("Invalid subscription id") from e

    if subscription:
        return subscription
    else:
        raise ValueError(f"Subscription with {subscription_id} does not exist in the database")


def update_subscription_status(subscription_id: UUIDstr, status: str) -> SubscriptionTable:
    """Update the subscription status.

    Args:
        subscription_id: Id of the subscription to update.
        status: Status to transition to.

    Returns: Subscription

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = status
    return subscription


def update_subscription_description(subscription_id: UUIDstr, description: str) -> SubscriptionTable:
    """Update a subscriptions description.

    Args:
        subscription_id: subscription id of the subscription to update
        description: new subscription description

    Returns: Subscription

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.description = description
    return subscription


def activate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Activate subscription by subscription id.

    Args:
        subscription_id: subscription id of the subscription

    Returns: Subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "active"
    subscription.start_date = nowtz()
    subscription.insync = True
    return subscription


def provision_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Provision subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "provisioning"
    subscription.insync = True
    return subscription


def migrate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Migrate subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "migrating"
    subscription.insync = False
    return subscription


def unsync(subscription_id: UUIDstr, checked: bool = True) -> SubscriptionTable:
    """Unsync subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    if checked and not subscription.insync:
        raise ValueError("Subscription is already out of sync, cannot continue!")
    subscription.insync = False
    return subscription


def resync(subscription_id: UUIDstr) -> SubscriptionTable:
    """
    Resync subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.insync = True
    return subscription


def set_status_provisioning_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    return update_subscription_status(subscription_id, "provisioning")


def terminate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    return update_subscription(subscription_id, status="terminated", end_date=nowtz())


def create_subscription(
    organisation: UUIDstr, product: ProductTable, subscription_name: str, subscription_id: UUIDstr
) -> UUID:
    subscription = SubscriptionTable(
        subscription_id=subscription_id,
        product_id=product.product_id,
        customer_id=organisation,
        description=subscription_name,
        start_date=None,
        end_date=None,
        insync=False,
        status="initial",
    )
    db.session.add(subscription)

    return subscription.subscription_id


def update_subscription(subscription_id: str, **attrs: Union[Dict, UUIDstr, str, datetime]) -> SubscriptionTable:
    """
    Update the subscription.

    Args:
        subscription_id: SubscriptionTable id of the subscription
        attrs: Attributes that will be set

    Returns: Subscription

    """

    subscription = get_subscription(subscription_id, for_update=True)

    for (key, value) in attrs.items():
        setattr(subscription, key, value)

    return subscription


def retrieve_node_subscriptions_by_name(node_name: str) -> List[SubscriptionTable]:
    node_subscriptions = (
        SubscriptionTable.query.join(
            ProductTable, SubscriptionInstanceTable, SubscriptionInstanceValueTable, ResourceTypeTable
        )
        .filter(SubscriptionInstanceValueTable.value == node_name)
        .filter(ResourceTypeTable.resource_type == "nso_device_id")
        .filter(SubscriptionTable.status.in_(["active", "provisioning"]))
        .all()
    )
    return node_subscriptions


def retrieve_subscription_by_subscription_instance_value(
    resource_type: str, value: str, sub_status: Tuple = ("provisioning", "active")
) -> Optional[SubscriptionTable]:
    """
    Retrieve a Subscriptions by resource_type and value.

    Args:
        resource_type: name of the resource type
        value: value of the resource type
        sub_status: status of the subscriptions

    Returns: Subscription or None

    """
    subscription = (
        SubscriptionTable.query.join(SubscriptionInstanceTable, SubscriptionInstanceValueTable, ResourceTypeTable)
        .filter(SubscriptionInstanceValueTable.value == value)
        .filter(ResourceTypeTable.resource_type == resource_type)
        .filter(SubscriptionTable.status.in_(sub_status))
        .one_or_none()
    )
    return subscription


def find_values_for_resource_types(
    subscription_id: Union[UUID, UUIDstr], resource_types: Sequence[str], strict: bool = True
) -> Dict[str, List[str]]:
    """Find values for resource types by subscription ID.

    This function issues a single SQL query to find one or more resource types for a given subscription. As
    multiple values per resource type are possible (think of a BGP subscription with multiple SAPs, with each SAP
    sharing the same resource types) values are always returned as a list, even when there is only a single value.

    In case of shared resource types (hence multiple values), the order of the values matches across resource types:
    Meaning: in case of the above example with multiple, say two, SAPs, should we have requested the resource types::

        ('customer_ipv4_mtu', 'customer_ipv6_mtu', 'port_subscription_id'),

    we would get back::

        {
            'customer_ipv4_mtu': ['1500', '9000'],
            'customer_ipv6_mtu': ['9000', '9000'],
            'port_subscription_id': ['654a1e7f-9afb-43b7-ba8f-cdf9d48494aa', '2f86be06-a5b7-40ee-a71b-137d6b48a37e']
        }

    The first element of each list belong to the same SAP, likewise the second element of each list belongs to the
    same SAP. This allows you to implicitly correlate values.

    Args:
        subscription_id: The id of the subscription.
        resource_types: A sequence of resource type names.
        strict:
            True: raise `ValueError` if one or more requested resource types were not found,
            False: ignore resource types that weren't found.

    Returns:
        A dictionary of resource type names to lists of values.

    Raises:
        ValueError: if strict == True and one or more resource types were requested but not found.

    """
    # the `order_by` on `subscription_instance_id` is there to guarantee the matched ordering across resource_types
    # (see also docstring)
    query_result = (
        SubscriptionInstanceValueTable.query.join(ResourceTypeTable, SubscriptionInstanceTable)
        .filter(
            SubscriptionInstanceTable.subscription_id == subscription_id,
            ResourceTypeTable.resource_type.in_(resource_types),
        )
        .order_by(SubscriptionInstanceTable.subscription_instance_id)
        .values(ResourceTypeTable.resource_type, SubscriptionInstanceValueTable.value)
    )
    resource_type_values = tuple(query_result)

    rt2v: Dict[str, List[str]] = defaultdict(list)
    for resource_type, value in resource_type_values:
        rt2v[resource_type].append(value)
    if strict:
        missing = set(resource_types) - set(rt2v.keys())
        if missing:
            raise ValueError(f"Could not find requested resource types: '{','.join(missing)}'!")
    return rt2v


def query_parent_subscriptions(subscription_id: UUID) -> Query:
    """
    Return a query with all subscriptions -parents- that use this subscription_id with resource_type.

    The query can be used to add extra filters when/where needed.
    """
    # Find relations through resource types
    resource_type_relations = (
        SubscriptionTable.query.join(SubscriptionInstanceTable)
        .options(joinedload("customer_descriptions"))
        .join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(SubscriptionInstanceValueTable.value == str(subscription_id))
        .with_entities(SubscriptionTable.subscription_id)
    )

    # Find relations through instance hierarchy
    parent_instances = aliased(SubscriptionInstanceTable)
    child_instances = aliased(SubscriptionInstanceTable)
    relation_relations = (
        SubscriptionTable.query.join(parent_instances.subscription)
        .join(parent_instances.children_relations)
        .join(child_instances, SubscriptionInstanceRelationTable.child)
        .filter(child_instances.subscription_id == subscription_id)
        .filter(parent_instances.subscription_id != subscription_id)
        .with_entities(SubscriptionTable.subscription_id)
    )

    return SubscriptionTable.query.filter(
        or_(
            SubscriptionTable.subscription_id.in_(resource_type_relations.scalar_subquery()),
            SubscriptionTable.subscription_id.in_(relation_relations.scalar_subquery()),
        )
    )


def query_child_subscriptions(subscription_id: UUID) -> Query:
    """
    Return a query with all subscriptions -children- that are used by this subscription in resource_type.

    The query can be used to add extra filters when/where needed.
    """
    # Find relations through resource types
    resource_type_relations = (
        SubscriptionInstanceTable.query.join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(SubscriptionInstanceTable.subscription_id == subscription_id)
        .join(SubscriptionTable, SubscriptionInstanceValueTable.value == cast(SubscriptionTable.subscription_id, Text))
        .with_entities(SubscriptionTable.subscription_id)
    )

    # Find relations through instance hierarchy
    parent_instances = aliased(SubscriptionInstanceTable)
    child_instances = aliased(SubscriptionInstanceTable)
    relation_relations = (
        SubscriptionTable.query.join(child_instances.subscription)
        .join(child_instances.parent_relations)
        .join(parent_instances, SubscriptionInstanceRelationTable.parent)
        .filter(parent_instances.subscription_id == subscription_id)
        .filter(child_instances.subscription_id != subscription_id)
        .with_entities(SubscriptionTable.subscription_id)
    )

    return SubscriptionTable.query.filter(
        or_(
            SubscriptionTable.subscription_id.in_(resource_type_relations.scalar_subquery()),
            SubscriptionTable.subscription_id.in_(relation_relations.scalar_subquery()),
        )
    )


def _terminated_filter(query: Query) -> List[UUID]:
    return list(
        more_itertools.flatten(
            query.filter(SubscriptionTable.status != "terminated").with_entities(SubscriptionTable.subscription_id)
        )
    )


def _in_sync_filter(query: Query) -> List[UUID]:
    return list(
        more_itertools.flatten(
            query.filter(not_(SubscriptionTable.insync)).with_entities(SubscriptionTable.subscription_id)
        )
    )


RELATION_RESOURCE_TYPES: List[str] = []


def status_relations(subscription: SubscriptionTable) -> Dict[str, List[UUID]]:
    """Return info about locked children or parents.

    This call will be used by the client to determine if it's safe to
    start a modify or terminate workflow. There are 4 cases:

    1) The subscription is a IP, LightPath or ELAN: the child subscriptions are checked for not 'insync' instances.
    2) The subscription is a ServicePort: parent subscription are checked for not 'insync' instances and for parent
       services that are not terminated.
    3) The subscription is a Node: Related Core link subscriptions are checked that there are no active instances
       This is only used for the terminate workflow and ignored for modify
    4) IP_prefix cannot be terminated when in use

    """
    parent_query = query_parent_subscriptions(subscription.subscription_id)

    unterminated_parents = _terminated_filter(parent_query)
    locked_parent_relations = _in_sync_filter(parent_query)

    query = query_child_subscriptions(subscription.subscription_id)

    locked_child_relations = _in_sync_filter(query)

    result = {
        "locked_relations": locked_parent_relations + locked_child_relations,
        "unterminated_parents": unterminated_parents,
    }

    logger.debug(
        "Returning status info for related subscriptions",
        result=result,
        subscription_id=str(subscription.subscription_id),
    )
    return result


def get_relations(subscription_id: UUIDstr) -> Dict[str, List[UUID]]:
    subscription_table = SubscriptionTable.query.options(joinedload("product"), joinedload("product.workflows")).get(
        subscription_id
    )
    relations = status_relations(subscription_table)
    return relations


TARGET_DEFAULT_USABLE_MAP: Dict[Target, List[str]] = {
    Target.CREATE: [],
    Target.MODIFY: ["active"],
    Target.TERMINATE: ["active", "provisioning"],
    Target.SYSTEM: ["active"],
}

WF_USABLE_MAP: Dict[str, List[str]] = {}

WF_BLOCKED_BY_PARENTS: Dict[str, bool] = {}


def subscription_workflows(subscription: SubscriptionTable) -> Dict[str, Any]:
    """
    Return a dict containing all the workflows a user can start for this subscription.

    Args:
        subscription: an SqlAlchemy instance of a `db.SubscriptionTable`

    Returns:
        A dictionary with the following structure (reason and its related keys are only present when workflows are blocked):

        >>> {  # doctest:+SKIP
        ...     "reason": "Optional global reason like subscription is in use"
        ...     "create": [{"name": "workflow.name", "description": "workflow.description", "reason": "Optional reason why this specific workflow is blocked"}],
        ...     "modify": [],
        ...     "terminate": [],
        ...     "system": [],
        ... }

    """
    default_json: Dict[str, Any] = {}

    if not subscription.insync:
        default_json["reason"] = "subscription.not_in_sync"
    else:
        data = status_relations(subscription)

        if data["locked_relations"]:
            default_json["reason"] = "subscription.relations_not_in_sync"
            default_json["locked_relations"] = data["locked_relations"]

    workflows: Dict[str, Any] = {
        "create": [],
        "modify": [],
        "terminate": [],
        "system": [],
    }

    for workflow in subscription.product.workflows:
        if workflow.name in ["modify_note"] or workflow.target == Target.SYSTEM:
            # validations and modify note are also possible with: not in sync or locked relations
            workflow_json = {"name": workflow.name, "description": workflow.description}
        else:
            workflow_json = {"name": workflow.name, "description": workflow.description, **default_json}

        # Simple reasons like "not in sync" and "locked_relations" are handled now. Handle special lifecycle cases:
        if "reason" not in workflow_json:

            default = TARGET_DEFAULT_USABLE_MAP[workflow.target]
            usable_when = WF_USABLE_MAP.get(workflow.name, default)

            if subscription.status not in usable_when:
                workflow_json["reason"] = "subscription.no_modify_invalid_status"
                workflow_json["usable_when"] = usable_when
                workflow_json["status"] = subscription.status
                workflow_json["action"] = "terminated" if workflow.target == Target.TERMINATE else "modified"

            # Check if this workflow is blocked because there are unterminated relations
            blocked_by_parents = WF_BLOCKED_BY_PARENTS.get(workflow.name, workflow.target == Target.TERMINATE)
            if blocked_by_parents and data["unterminated_parents"]:
                workflow_json["reason"] = "subscription.no_modify_parent_subscription"
                workflow_json["unterminated_parents"] = data["unterminated_parents"]
                workflow_json["action"] = "terminated" if workflow.target == Target.TERMINATE else "modified"

        workflows[workflow.target.lower()].append(workflow_json)

    return {**workflows, **default_json}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Protocol, cast

from schedule import CancelJob


class SchedulingFunction(Protocol):
    __name__: str
    name: str
    time_unit: str
    period: Optional[int]
    at: Optional[str]

    def __call__(self) -> Optional[CancelJob]:
        ...


def scheduler(
    name: str, time_unit: str, period: int = 1, at: Optional[str] = None
) -> Callable[[Callable[[], Optional[CancelJob]]], SchedulingFunction]:
    """Create schedule.

    Either specify the period or the at. Examples:
    time_unit = "hours", period = 12 -> will run every 12 hours
    time_unit = "day", at="01:00" -> will run every day at 1 o'clock
    """

    def _scheduler(f: Callable[[], Optional[CancelJob]]) -> SchedulingFunction:
        schedule = cast(SchedulingFunction, f)
        schedule.name = name
        schedule.time_unit = time_unit
        schedule.period = period
        schedule.at = at
        return schedule

    return _scheduler



from typing import Optional

import pytest

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_sub_block_one():
    class SubBlockOneForTestInactive(ProductBlockModel, product_block_name="SubBlockOneForTest"):
        int_field: Optional[int] = None
        str_field: Optional[str] = None

    class SubBlockOneForTestProvisioning(SubBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        int_field: int
        str_field: Optional[str] = None

    class SubBlockOneForTest(SubBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        int_field: int
        str_field: str

    return SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest


@pytest.fixture
def test_product_sub_block_one_db(resource_type_int, resource_type_str):
    sub_block = ProductBlockTable(
        name="SubBlockOneForTest", description="Test Sub Block One", tag="TEST", status="active"
    )

    sub_block.resource_types = [resource_type_int, resource_type_str]

    db.session.add(sub_block)
    db.session.commit()
    return sub_block
